#!/usr/bin/env python

#####################################
# Rashba_Exciton.py version 1.0
# Author and maintainer: Michael W. Swift <michael.swift.ctr@nrl.navy.mil>
# Co-authors: John L. Lyons <john.lyons@nrl.navy.mil>
#             Alexander L. Efros <sasha.efros@nrl.navy.mil>
#             Peter C. Sercel <psercel@caltech.edu>

# This code implements the effective-mass model of a Rashba Exciton as presented in:
# M. W. Swift, J. L. Lyons, Al. L. Efros, and P. C. Sercel, "Rashba exciton in a 2D 
# perovskite quantum dot" Nanoscale (2021), DOI: 10.1039/d1nr04884h.
#####################################

import numpy as np
from numpy.lib.scimath import sqrt
import scipy as sp
from scipy.special import j0, j1, jv
from scipy.optimize import brentq, minimize_scalar
from scipy.integrate import quad, quad_vec, tplquad
from scipy import optimize 

import sys
from matplotlib import pyplot as plt
from matplotlib.cm import get_cmap
from integrandLR import integrandLR
import yaml
from time import process_time
try:
    import vegas
except:
    print("WARNING: Failed to import vegas.  Will not be able to perform LR integrals")

a0 = 0.0529177211 # bohr radius, nm
c = 0.0262470125962755 # c=2m_o/hbar^2, meV^-1 nm^-2

class Rashba_Exciton:

    def __init__(self, input_fname = "Rashba_exciton_options.yaml", input_dict = None):
            
        if input_dict is None:
            with open(input_fname, "r") as fin:
                input_dict = yaml.load(fin, Loader=yaml.FullLoader)

        self.w        = float(input_dict["w"])
        self.d        = float(input_dict["L"]) # Note d is called \mathcal{L} in the paper
        self.Lambda   = float(input_dict["Lambda"])
        self.Omega    = float(input_dict["Omega"])
        self.epsiloni = float(input_dict["epsiloni"])
        self.epsilono = float(input_dict["epsilono"])
        ERashba_e     = float(input_dict["ERashba_e"])
        ERashba_h     = float(input_dict["ERashba_h"])
        self.mu       = float(input_dict["mu"])
        self.Egap     = float(input_dict["Egap"])
        self.f_para   = float(input_dict["f_para"])
        self.f_perp   = float(input_dict["f_perp"])
        theta         = float(input_dict["theta"])
        self.Fz       = int(input_dict["Fz"])
        self.tol      = float(input_dict["tol"])

        self.rootsearch_dE = float(input_dict["rootsearch_dE"])
        internal_Rashba_flag = bool(input_dict["internal_Rashba_flag"])

        self.wf_tol = self.tol
        if "wf_tol" in input_dict:
            self.wf_tol = float(input_dict["wf_tol"])

        Rashba_sign = -1
        if "Rashba_sign" in input_dict:
            Rashba_sign   = int(input_dict["Rashba_sign"])


        # Dielectric confinement
        # Calculate Ar, the effective 2D exciton radius, and the binding energy
        # Recommend to pre-calculate these parameters to avoid repeating this step.
        if all(key in input_dict for key in ["ArL", "a2D", "Ebind"]):
            Ar  = input_dict["ArL"]
            self.a2D = input_dict["a2D"]
            self.Ebind = input_dict["Ebind"]
        else:
            print("Calculating dielectric confinement effects...")
            print("Note that this routine uses a stochastic integration method.")
            print("To save time in the future and ensure precise reproducibility,")
            print("I recommend setting the results in the input file for future runs.")
            Ar, self.a2D, self.Ebind = self.dielectric_confinement()

            print()
            print("Dielectric confinement calculation complete.  Results:")
            print(f"ArL       : {Ar:5.3f}   #        geometric Rashba factor, 2D value 0.105")
            print(f"a2D       : {self.a2D:5.3f}   # nm,    2D exciton radius")
            print(f"Ebind     : {self.Ebind:8.3f} # meV,  Exciton binding energy (<0: bound)")


        # Calculate derived quantities
        w  = self.w
        mu = self.mu
        me = 2*mu
        mh = 2*mu
        ae = sqrt(4/c * ERashba_e/me)
        ah = Rashba_sign*sqrt(4/c * ERashba_h/mh)
        E_Rashba = ae*ah*mu * c/2 # Rashba exchange energy \mathcal{E}_R, meV

        self.M = me + mh # COM mass
        self.ap = ae*me + ah*mh # alpha^{+}_{ex}
        self.am = ae*me - ah*mh # alpha^{-}_{ex}
        self.Ep = 3*self.Egap/sqrt(me*mh) # Kane parameter, meV
        
        # k=0 splitting
        self.Ed = 0 - 4*Ar*E_Rashba*internal_Rashba_flag 
        self.Et = w*(np.cos(theta))**2
        self.Ez = 2*w*(np.sin(theta))**2 + 4*Ar*E_Rashba*internal_Rashba_flag

        # Find dispersion minima
        res0 = minimize_scalar(lambda k: self.E0minus(k**2), bounds=(-1.5,1.5))
        res1 = minimize_scalar(lambda k: self.E1minus(k**2), bounds=(-1.5,1.5))
        self.E0_disp_min = res0.fun
        self.E1_disp_min = res1.fun
        if self.E0_disp_min < self.E1_disp_min:
            self.E_disp_min = self.E0_disp_min
            self.ksq_disp_min = (res0.x)**2
        else:
            self.E_disp_min = self.E1_disp_min
            self.ksq_disp_min = (res1.x)**2

        # 3D bohr radius
        self.ax = a0*self.epsiloni/mu
        # Internal wavefunction using 2D radius
        self.phi10sq = 8/(np.pi * self.a2D**2)

        # LR exchange
        self.eps_avg = self.epsiloni*self.d/self.Lambda + self.epsilono*(1-self.d/self.Lambda)
        self.hbar_omega_LT = 6.89421148e-9 * self.f_para / (self.eps_avg * self.Egap * self.Lambda)
                                                                       # Prefactor from LR_Exchange_Integral.nb

        self.g0 = sqrt(2) * np.sin(theta)
        self.g1 = np.cos(theta)



    # Define dispersion functions (Eq. 16)
    def E0minus(self, ksq):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return 1/c * 1/M * ksq + ((Ed+Et)*M - sqrt((4*am**2 * ksq) + ((Ed-Et)**2 * M**2)))/(2*M)

    def E0plus(self, ksq):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return 1/c * 1/M * ksq + ((Ed+Et)*M + sqrt((4*am**2 * ksq) + ((Ed-Et)**2 * M**2)))/(2*M)

    def E1minus(self, ksq):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return 1/c * 1/M * ksq + ((Et+Ez)*M - sqrt((4*ap**2 * ksq) + ((Et-Ez)**2 * M**2)))/(2*M)

    def E1plus(self, ksq):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return 1/c * 1/M * ksq + ((Et+Ez)*M + sqrt((4*ap**2 * ksq) + ((Et-Ez)**2 * M**2)))/(2*M)


    # Eq S4
    def k11sq(self, E):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return E*M*c + 0.5*(ap**2 * c**2 - c*M*Et - c*M*Ez + \
            sqrt(c**2 * (ap**4*c**2 + 2*ap**2*c*(2*E-Et-Ez)*M + (Et-Ez)**2*M**2)) )

    def k12sq(self, E):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return E*M*c + 0.5*(ap**2 * c**2 - c*M*Et - c*M*Ez - \
            sqrt(c**2 * (ap**4*c**2 + 2*ap**2*c*(2*E-Et-Ez)*M + (Et-Ez)**2*M**2)) )

    def k01sq(self, E):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return E*M*c + 0.5*(am**2 * c**2 - c*M*Ed - c*M*Et + \
            sqrt(c**2 * (am**4*c**2 + 2*am**2*c*(2*E-Ed-Et)*M + (Ed-Et)**2*M**2)) )

    def k02sq(self, E):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return E*M*c + 0.5*(am**2 * c**2 - c*M*Ed - c*M*Et - \
            sqrt(c**2 * (am**4*c**2 + 2*am**2*c*(2*E-Ed-Et)*M + (Ed-Et)**2*M**2)) )

    # Eq (S2)
    def c1plus(self, ksq):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return (Et*M - Ez*M + sqrt(4*ap**2*ksq + (Et-Ez)**2*M**2))/(2*ap*sqrt(ksq))

    def c1minus(self, ksq):
        M = self.M
        ap = self.ap
        Et = self.Et
        Ez = self.Ez
        return (Et*M - Ez*M - sqrt(4*ap**2*ksq + (Et-Ez)**2*M**2))/(2*ap*sqrt(ksq))

    # Eq (S3)
    def c0plus(self, ksq):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return (-Ed*M + Et*M - sqrt(4*am**2*ksq + (Ed-Et)**2*M**2))/(2*am*sqrt(ksq))

    def c0minus(self, ksq):
        M = self.M
        am = self.am
        Ed = self.Ed
        Et = self.Et
        return (-Ed*M + Et*M + sqrt(4*am**2*ksq + (Ed-Et)**2*M**2))/(2*am*sqrt(ksq))


    # find ksquared solutions for a given energy E and assign them to branches (Eqs (22) and (S7))
    def get_ksq_and_c(self, E):

        my_k11sq=self.k11sq(E)
        my_k12sq=self.k12sq(E)
        my_k01sq=self.k01sq(E)
        my_k02sq=self.k02sq(E)

        if np.abs(self.E1minus(my_k11sq) - E) < self.tol:
            c11=self.c1minus(my_k11sq)
            branch11="E1minus"
        else:
            if np.abs(self.E1plus(my_k11sq) - E) > self.tol:
                print(f"WARNING: k11sq={my_k11sq:5f} and E={E} and does not match E1minus or E1plus:")
                print(f"   E1plus(my_k11sq) - E = {self.E1plus(my_k11sq) - E}")
                print(f"   E1minus(my_k11sq)- E = {self.E1minus(my_k11sq) - E}")
                print(f"   Assigning to E1plus")
            c11=self.c1plus(my_k11sq)
            branch11="E1plus"

        if np.abs(self.E1minus(my_k12sq) - E) < self.tol:
            c12=self.c1minus(my_k12sq)
            branch12="E1minus"
        else:
            if np.abs(self.E1plus(my_k12sq) - E) > self.tol:
                print(f"WARNING: k12sq={my_k12sq:5f} and E={E} does not match E1minus or E1plus:")
                print(f"   E1plus(my_k12sq) - E = {self.E1plus(my_k12sq) - E}")
                print(f"   E1minus(my_k12sq)- E = {self.E1minus(my_k12sq) - E}")
                print(f"   Assigning to E1plus")
            c12=self.c1plus(my_k12sq)
            branch12="E1plus"

        if np.abs(self.E0minus(my_k01sq) - E) < self.tol:
            c01=self.c0minus(my_k01sq)
            branch01="E0minus"
        else:
            if np.abs(self.E0plus(my_k01sq) - E) > self.tol:
                print(f"WARNING: k01sq={my_k01sq:5f} and E={E} does not match E0minus or E0plus:")
                print(f"   E0plus(my_k01sq) - E = {self.E0plus(my_k01sq) - E}")
                print(f"   E0minus(my_k01sq)- E = {self.E0minus(my_k01sq) - E}")
                print(f"   Assigning to E0plus")
            c01=self.c0plus(my_k01sq)
            branch01="E0plus"

        if np.abs(self.E0minus(my_k02sq) - E) < self.tol:
            c02=self.c0minus(my_k02sq)
            branch02="E0minus"
        else:
            if np.abs(self.E0plus(my_k02sq) - E) > self.tol:
                print(f"WARNING: k02sq={my_k02sq:5f} and E={E} does not match E0minus or E0plus:")
                print(f"   E0plus(my_k02sq) - E = {self.E0plus(my_k02sq) - E}")
                print(f"   E0minus(my_k02sq)- E = {self.E0minus(my_k02sq) - E}")
                print(f"   Assigning to E0plus")
            c02=self.c0plus(my_k02sq)
            branch02="E0plus"

        return ((my_k11sq, c11, branch11), \
                (my_k12sq, c12, branch12), \
                (my_k01sq, c01, branch01), \
                (my_k02sq, c02, branch02))

    # Calculate dielectric confinement (Hanamura method), Eqs. (5), (6)
    def dielectric_confinement(self, nnMax=8):

        kineticPrefactor = 38.0998212/self.mu
        prefactorHan = 1439.964548/self.epsiloni
        kappa = self.epsiloni/self.epsilono
        d = self.d
        nnMax=8

        def HanKinetic10(a, r, ze, zh):
            
            return kineticPrefactor * (64*(a - 2*r)*np.power(np.cos((np.pi*ze)/d),2)*np.power(np.cos((np.pi*\
                    zh)/d),2))/(np.power(a,4)*np.power(d,2)*np.pi*r)*np.exp((-4*r)/a)

        def HanKinetic21(a, r, ze, zh):
            return kineticPrefactor * (512*(9*a - \
                    2*r)*r*np.power(np.cos((np.pi*ze)/d),2)*np.power(np.cos((np.pi*zh)/d),\
                    2))/(2187.*np.power(a,6)*np.power(d,2)*np.pi)*np.exp((-4*r)/(3.*a))

        def Znh(nn, zh):
            return ((-1)**np.abs(nn))*zh + d*nn

        def q(nn):
            return ((kappa-1)/(kappa+1))**(np.abs(nn))

        def HanV(rho, ze, zh):

            zeroTerm = 1/sqrt(rho**2 + (ze-zh)**2)
            nnTerm = lambda nn: q(nn)/sqrt(rho**2 + (ze-Znh(nn,zh))**2) \
                            + q(-nn)/sqrt(rho**2 + (ze-Znh(-nn,zh))**2)

            return prefactorHan*(zeroTerm + np.sum([nnTerm(nn) for nn in range(1, nnMax+1)],axis=0))

        def HanPotential10(a, r, ze, zh):
            return (2/d * np.cos((np.pi*ze)/d) * np.cos((np.pi*zh)/d))**2 * \
                    HanV(r, ze, zh) * (-8 * np.exp(-4*r/a)/(a**2 * np.pi))

        def HanPotential21(a, r, ze, zh):
            return (2/d * np.cos((np.pi*ze)/d) * np.cos((np.pi*zh)/d))**2 * \
                    HanV(r, ze, zh) * (-64 * np.exp(-4*r/(3*a))*r**2/(243* a**4 * np.pi))

        def HanEnergy10(a, verbose=False, nitn_train=10, neval_train=1e5, 
            nitn_final=10, neval_final=1e5):

            verboseprint = print if verbose else lambda *a, **k: None

            t_start=process_time()

            def integrand(xvec):
                x = xvec[:,0]
                ze = xvec[:,1]
                zh = xvec[:,2]
                
                r_scale=1
                
                r=r_scale*x/(1-x)
                jacobian = r_scale/(1-x) + r_scale*x/((1-x)**2)
                return r * jacobian *\
                        (HanKinetic10(a, r, ze, zh) + HanPotential10(a, r, ze, zh))

            @vegas.batchintegrand
            def f_batch(x):
                return integrand(x)

            integ = vegas.Integrator([[0,1], [-self.d/2, self.d/2], [-self.d/2, self.d/2]])
            verboseprint("Training runs in progress...")
            integ(f_batch, nitn=nitn_train, neval=neval_train)
            t_intermediate = process_time()
            verboseprint(f"Training runs complete.  Elapsed time: {t_intermediate-t_start} sec.")
            verboseprint("Final runs in progress...")
            result = integ(f_batch, nitn=nitn_final, neval=neval_final)
            t_stop=process_time()

            verboseprint(result.summary())
            verboseprint()
            verboseprint('result = %s   Q = %.2f' % (result, result.Q))
            verboseprint(f'elapsed time: {t_stop-t_start} sec.')
            

            return 2*np.pi * result.mean

        def HanEnergy21(a, verbose=False, nitn_train=10, neval_train=1e5, 
            nitn_final=10, neval_final=1e5):

            verboseprint = print if verbose else lambda *a, **k: None

            t_start=process_time()

            def integrand(xvec):
                x = xvec[:,0]
                ze = xvec[:,1]
                zh = xvec[:,2]
                
                r_scale=1
                
                r=r_scale*x/(1-x)
                jacobian = r_scale/(1-x) + r_scale*x/((1-x)**2)

                return r * jacobian * \
                        (HanKinetic21(a, r, ze, zh) + HanPotential21(a, r, ze, zh))

            @vegas.batchintegrand
            def f_batch(x):
                return integrand(x)

            integ = vegas.Integrator([[0,1], [-self.d/2, self.d/2], [-self.d/2, self.d/2]])
            verboseprint("Training runs in progress...")
            integ(f_batch, nitn=nitn_train, neval=neval_train)
            t_intermediate = process_time()
            verboseprint(f"Initial runs complete.  Elapsed time: {t_intermediate-t_start} sec.")
            verboseprint("Final runs in progress...")
            result = integ(f_batch, nitn=nitn_final, neval=neval_final)
            t_stop=process_time()

            verboseprint(result.summary())
            verboseprint()
            verboseprint('result = %s   Q = %.2f' % (result, result.Q))
            verboseprint(f'elapsed time: {t_stop-t_start} sec.')
            
            return 2*np.pi * result.mean

        res10 = optimize.minimize_scalar(HanEnergy10, bracket=(1,4.5,10))
        a10 = res10.x
        E10 = res10.fun

        res21 = optimize.minimize_scalar(HanEnergy21, bracket=(0.5,1.0,8.0))
        a21 = res21.x
        E21 = res21.fun

        RydXMeV = 1/c * 1/self.mu * 1/(a10)**2

        calculated_Ar = 4 * a10**2 * (384*a10**2 * a21**2)/(a10 + 3*a21)**6 * RydXMeV / (E21 - E10)

        return calculated_Ar, a10, E10


    # Eq (25)
    def systemMat(self,r,k11,k12,k01,k02,c11,c12,c01,c02):
        Fz = self.Fz
        return np.array([
            [0                  , 0                   , c01*jv(Fz, k01*r), c02*jv(Fz, k02*r)],
            [-c11*jv(Fz-1,k11*r), -c12*jv(Fz-1,k12*r) , jv(Fz-1, k01*r)  , jv(Fz-1, k02*r)  ],
            [jv(Fz,k11*r)       , jv(Fz,k12*r)        , 0                , 0                ],
            [c11*jv(Fz+1,k11*r) , c12*jv(Fz+1,k12*r)  , jv(Fz+1, k01*r)  , jv(Fz+1, k02*r)  ]])


    # Solve Eq (25)
    def confinement(self, E, R, return_det=True):
        ((my_k11sq, c11, _), \
         (my_k12sq, c12, _), \
         (my_k01sq, c01, _), \
         (my_k02sq, c02, _)) = self.get_ksq_and_c(E)

        k11 = sqrt(my_k11sq)
        k12 = sqrt(my_k12sq)
        k01 = sqrt(my_k01sq)
        k02 = sqrt(my_k02sq)

        systemMatR = self.systemMat(R,k11,k12,k01,k02,c11,c12,c01,c02)

        
        if return_det:
            # A bit of a kludge: if the determinant is imaginary, return the imaginary part.
            # Else, return the real part
            det = np.linalg.det(systemMatR)
            if np.abs(np.real(det)) < self.tol:
                return np.imag(det)
            else:
                return np.real(det)

        else:

            # Need larger relative condition number at higher energies
            rcond = self.tol * max(E - self.E_disp_min,1)

            try:
                ABCD = sp.linalg.null_space(systemMatR, rcond=rcond)
            except ValueError as err:
                print(f"sp.linalg.null_space encountered ValueError at energy {E}:")
                print(err)
                return (None, None)

            if ABCD.shape[1] == 0:
                # Fake solution

                return (None, None)
            elif ABCD.shape[1] > 1:
                print(f"Warning!  Dimension of nullspace at energy {E} is {ABCD.shape[1]}.  Using the first basis vector.")
            
            systemParams = np.array((k11,k12,k01,k02,c11,c12,c01,c02))

            return (ABCD[:,0], systemParams)



    def build_wf(self, E, R, return_integrandLR=False):
        Fz = self.Fz
        g0 = self.g0
        g1 = self.g1
        d = self.d
        (ABCD, systemParams) = self.confinement(E, R, return_det=False)
        if ABCD is None:
            print(f"Fake solution discarded at E={E:5f} (dimension of nullspace is 0)")
            return

        wf_norm_integrand = lambda r: r * np.dot(self.systemMat(r, *systemParams), ABCD)**2
        N = 1/sqrt(np.sum(quad_vec(wf_norm_integrand, 0, R)[0]))
        wf = lambda r: N * np.dot(self.systemMat(r, *systemParams), ABCD)
        
        if return_integrandLR:
            my_integrandLR = lambda x: integrandLR(x, N, Fz, g0, g1, R, d, ABCD, *systemParams)
            return wf, my_integrandLR
        else:
            return wf


    def plot_wf(self, wf, R, fname="wf.pdf", return_plt=False):
        Fz = self.Fz
        rdat = np.linspace(0,R,100)
        wfdat = [wf(r) for r in rdat]
        plt.figure()
        plt.plot(rdat,np.real(wfdat))
        plt.gca().set_prop_cycle(None)
        plt.plot(rdat,np.imag(wfdat),"--", zorder=0)
        plt.xlabel("$\\rho$ (nm)", size=14)
        plt.ylabel("$\\Psi$", size=14)
        plt.legend([f'$|0,0\\rangle|{Fz}\\rangle$', f'$|1,+1\\rangle|{Fz-1}\\rangle$', 
                    f'$|1,0\\rangle|{Fz}\\rangle$', f'$|1,-1\\rangle|{Fz+1}\\rangle$'],
                    fontsize=14, loc="center right", bbox_to_anchor=(1.45,0.5))
        plt.xlim([0,R])

        if fname is not None:
            plt.savefig(fname, bbox_inches="tight")

        if return_plt:
            return plt
        else:
            plt.close()
            return

    # See Section 3 (oscillator transition strength) and ESI Section V
    def oscillator_strength(self, wf, R, E):

        # Note this returns oscillator strength per nm^2
        # Mutliply by np.pi * R**2 for total oscillator strength

        Fz = self.Fz
        a2D = self.a2D
        g0 = self.g0
        g1 = self.g1
        f_para = self.f_para
        phi10sq = self.phi10sq

        if abs(Fz)>1:
            return 0

        # Now Fz = 0 or +-1. The active row is at index 2-Fz
        # MWS: Corrected from 2+Fz, 2021/10/12   
        integrand = lambda x: x*wf(x*R)[2-Fz]
        radialOverlap = R**2 * quad_vec(integrand, 0, 1)[0]
        Ep_over_hbar_omega = np.pi * a2D**2/8 * 1/g1**2 * f_para*1.0e-14

        Ksq_per_unit_area = 2/(R**2) * np.abs(radialOverlap)**2 * phi10sq

        if Fz == 0:
            crystal_field_factor =  g0**2 
        else:
            crystal_field_factor = g1**2 


        return crystal_field_factor * Ksq_per_unit_area * Ep_over_hbar_omega


    def find_levels(self, R, Emin, Emax, return_osc=False):

        levels=[]
        wf_list=[]
        osc_list=[]

        determinant = lambda E: self.confinement(E, R)

        while 1:
            E1,E2 = rootsearch(determinant,Emin,Emax,self.rootsearch_dE)
            # print(E1,E2)
            if E1 != None:
                Emin = E2
                root = brentq(determinant, E1, E2)

                # Build the wavefunctions and test
                # (ABCD, systemParams) = confinement(root, R, return_det=False)
                wf = self.build_wf(root, R)

                if wf is None:
                    continue
                
                # Need looser tolerance at higher energies
                wf_tol = self.wf_tol * max(root - self.E_disp_min,1)

                if np.sum(np.abs(wf(R))) > wf_tol:
                    print(f"Fake solution discarded at E={root} (Phi(R) = {np.sum(np.abs(wf(R)))} > 0")
                else:
                    levels.append(root)
                    wf_list.append(wf)
                    if return_osc:
                        osc_list.append(self.oscillator_strength(wf, R, root))

            else:
                
                if return_osc:
                    return levels, wf_list, osc_list
                else:
                    return levels, wf_list

    # Long-range exchange calculation
    # See Section 4 and ESI Section VI
    def LR_exch(self, my_integrandLR, R, E, verbose=True, 
        nitn_train=10, neval_train=1e6, nitn_final=10, neval_final=1e6,
        n_selfcon_it=1000):

        d = self.d
        Egap = self.Egap
        Ebind = self.Ebind
        eps_avg = self.eps_avg
        epsiloni = self.epsiloni
        hbar_omega_LT = self.hbar_omega_LT
        Lambda = self.Lambda
        g1 = self.g1
        phi10sq = self.phi10sq
        tol=self.tol

        verboseprint = print if verbose else lambda *a, **k: None

        t_start=process_time()

        @vegas.batchintegrand
        def f_batch(x):
            return my_integrandLR(x)

        integ = vegas.Integrator(6 * [[0, 1]])
        verboseprint("Training runs in progress...")
        integ(f_batch, nitn=nitn_train, neval=neval_train)
        t_intermediate = process_time()
        verboseprint(f"Initial runs complete.  Elapsed time: {t_intermediate-t_start} sec.")
        verboseprint("Final runs in progress...")
        result = integ(f_batch, nitn=nitn_final, neval=neval_final)
        t_stop=process_time()

        verboseprint(result.summary())
        verboseprint()
        verboseprint('result = %s   Q = %.2f' % (result, result.Q))
        verboseprint(f'elapsed time: {t_stop-t_start} sec.')

        verboseprint(f'Discarding imaginary part {result[1]}, which should be zero')

        J_integral = self.phi10sq * result[0].mean
        verboseprint(f"J_Fz = {J_integral} nm^-3")

        hbar_omega_0 = Egap + Ebind + E
        # HLR = 1/(4*np.pi) * eps_avg/epsiloni * hbar_omega_LT * (Egap / hbar_omega)**2 \
                # * Lambda/(g1**2 * phi10sq) * J_integral
        HLR_func = lambda x: 1/(4*np.pi) * eps_avg/epsiloni * hbar_omega_LT * (Egap / (hbar_omega_0 + x))**2 \
                * Lambda/(g1**2 * phi10sq) * J_integral
        HLR = HLR_func(0)
        # verboseprint("Before selfconsistency:")
        # verboseprint(f"HLR / self.hbar_omega_LT = {HLR/self.hbar_omega_LT}")
        # verboseprint(f"HLR = {HLR} meV")

        nit=0

        while np.abs(HLR_func(HLR) - HLR) > tol and nit < n_selfcon_it :
            HLR = HLR_func(HLR)
            nit+=1

        if nit >= n_selfcon_it:
            print(f"WARNING: Long-range selfconsistency failed to converge after {nit} iterations.")

        verboseprint(f"HLR / self.hbar_omega_LT = {HLR/self.hbar_omega_LT}")
        verboseprint(f"HLR = {HLR} meV")

        return HLR


    # Produce plots like Figure 1.
    def plot_E_vs_k_func(self,xlim, ylim, xlim_aug_list=None, fname="E_vs_k.pdf", return_plt=False):
        kdat = np.linspace(xlim[0], xlim[1], 100)
        if xlim_aug_list is not None:
            for xlim_aug in xlim_aug_list:
                kdat_aug = np.linspace(xlim_aug[0], xlim_aug[1], 1000)
                kdat = np.concatenate((kdat, kdat_aug))
            kdat.sort()

        E0minus_dat = [self.E0minus(k**2) for k in kdat]
        E0plus_dat  = [self.E0plus (k**2) for k in kdat]
        E1minus_dat = [self.E1minus(k**2) for k in kdat]
        E1plus_dat  = [self.E1plus (k**2) for k in kdat]



        plt.figure()
        plt.plot([0,0], ylim, color="gray", lw=0.5)
        plt.plot(xlim, [0,0], color="gray", lw=0.5)
        plt.plot(kdat, E0minus_dat, label = "$E^{0-}$", color="black")
        plt.plot(kdat, E0plus_dat , label = "$E^{0+}$")
        plt.plot(kdat, E1minus_dat, label = "$E^{1-}$")
        plt.plot(kdat, E1plus_dat , label = "$E^{1+}$")
        plt.xlabel("$K$ (nm$^{-1}$)", size=14)
        plt.ylabel("$E$ (meV)", size=14)
        plt.legend(loc='center right', fontsize=14)
        plt.xlim(xlim)
        plt.ylim(ylim)

        if fname is not None:
            plt.savefig(fname, bbox_inches="tight")
        
        if return_plt:
            return plt
        else:
            plt.close()
            return

    # Produce plots like Figure 2
    def plot_E_vs_ksq_func(self, xlim, ylim, xlim_aug_list=None, fname="E_vs_ksq.pdf", return_plt=False):
        kdat = np.linspace(xlim[0], xlim[1], 100)
        if xlim_aug_list is not None:
            for xlim_aug in xlim_aug_list:
                kdat_aug = np.linspace(xlim_aug[0], xlim_aug[1], 1000)
                kdat = np.concatenate((kdat, kdat_aug))
            kdat.sort()
        E0minus_dat = [self.E0minus(ksq) if abs(np.imag(self.E0minus(ksq)))==0 else None for ksq in kdat ]
        E0plus_dat  = [self.E0plus (ksq) if abs(np.imag(self.E0plus (ksq)))==0 else None for ksq in kdat ]
        E1minus_dat = [self.E1minus(ksq) if abs(np.imag(self.E1minus(ksq)))==0 else None for ksq in kdat ]
        E1plus_dat  = [self.E1plus (ksq) if abs(np.imag(self.E1plus (ksq)))==0 else None for ksq in kdat ]

        plt.figure()
        plt.plot([0,0], ylim, color="gray", lw=0.5)
        plt.plot(xlim, [0,0], color="gray", lw=0.5)
        plt.plot(kdat, E0minus_dat, label = "$E^{0-}$", color="black")
        plt.plot(kdat, E0plus_dat , label = "$E^{0+}$")
        plt.plot(kdat, E1minus_dat, label = "$E^{1-}$")
        plt.plot(kdat, E1plus_dat , label = "$E^{1+}$")
        plt.xlabel("$K^2$ (nm$^{-2}$)", size=14)
        plt.ylabel("$E$ (meV)", size=14)
        plt.legend(fontsize=14)
        plt.xlim(xlim)
        plt.ylim(ylim)
        
        if fname is not None:
            plt.savefig(fname, bbox_inches="tight")

        if return_plt:
            return plt
        else:
            plt.close()
            return

    # Produce plots like Figure 2
    def plot_ksq_vs_E_func(self, xlim, ylim, xlim_aug_list=None, fname="ksq_vs_E.pdf", return_plt=False):
        
        cm = get_cmap('tab10')

        Edat = np.linspace(max(xlim[0],self.E_disp_min), xlim[1], 100)
        
        if xlim_aug_list is not None:
            for xlim_aug in xlim_aug_list:
                Edat_aug = np.linspace(max(xlim_aug[0],self.E_disp_min), xlim_aug[1], 1000)
                Edat = np.concatenate((Edat, Edat_aug))
            Edat.sort()

        E0minus_dat_1 = []
        E0minus_dat_2 = []
        E0plus_dat_1  = []
        E0plus_dat_2  = []
        E1minus_dat_1 = []
        E1minus_dat_2 = []
        E1plus_dat_1  = []
        E1plus_dat_2  = []

        for my_E in Edat:
            ((my_k11sq, c11, branch11), \
             (my_k12sq, c12, branch12), \
             (my_k01sq, c01, branch01), \
             (my_k02sq, c02, branch02)) = self.get_ksq_and_c(my_E)

            if branch11 == "E1minus":
                E1minus_dat_1.append(my_k11sq)
                E1plus_dat_1.append(None)
            else:
                E1plus_dat_1.append(my_k11sq)
                E1minus_dat_1.append(None)

            if branch12 == "E1minus":
                E1minus_dat_2.append(my_k12sq)
                E1plus_dat_2.append(None)
            else:
                E1plus_dat_2.append(my_k12sq)
                E1minus_dat_2.append(None)

            if branch01 == "E0minus":
                E0minus_dat_1.append(my_k01sq)
                E0plus_dat_1.append(None)
            else:
                E0plus_dat_1.append(my_k01sq)
                E0minus_dat_1.append(None)

            if branch02 == "E0minus":
                E0minus_dat_2.append(my_k02sq)
                E0plus_dat_2.append(None)
            else:
                E0plus_dat_2.append(my_k02sq)
                E0minus_dat_2.append(None)

        E0minus_dat_1 = np.array(E0minus_dat_1, dtype=complex)
        E0minus_dat_2 = np.array(E0minus_dat_2, dtype=complex)
        E0plus_dat_1  = np.array(E0plus_dat_1 , dtype=complex)
        E0plus_dat_2  = np.array(E0plus_dat_2 , dtype=complex)
        E1minus_dat_1 = np.array(E1minus_dat_1, dtype=complex)
        E1minus_dat_2 = np.array(E1minus_dat_2, dtype=complex)
        E1plus_dat_1  = np.array(E1plus_dat_1 , dtype=complex)
        E1plus_dat_2  = np.array(E1plus_dat_2 , dtype=complex)

        plt.figure()
        plt.plot([0,0], ylim, color="gray", lw=0.5)
        plt.plot(xlim, [0,0], color="gray", lw=0.5)

        if not np.all(np.isnan(E0minus_dat_1)):
            if np.nanmax(np.abs(np.imag(E0minus_dat_1))) > self.tol:
                plt.plot(Edat, np.imag(E0minus_dat_1), "--", color="black")
        if not np.all(np.isnan(E0plus_dat_1 )):
            if np.nanmax(np.abs(np.imag(E0plus_dat_1 ))) > self.tol:
                plt.plot(Edat, np.imag(E0plus_dat_1 ), "--", color=cm(0))
        if not np.all(np.isnan(E1minus_dat_1)):
            if np.nanmax(np.abs(np.imag(E1minus_dat_1))) > self.tol:
                plt.plot(Edat, np.imag(E1minus_dat_1), "--", color=cm(1))
        if not np.all(np.isnan(E1plus_dat_1 )):
            if np.nanmax(np.abs(np.imag(E1plus_dat_1 ))) > self.tol:
                plt.plot(Edat, np.imag(E1plus_dat_1 ), "--", color=cm(2))
        if not np.all(np.isnan(E0minus_dat_2)):
            if np.nanmax(np.abs(np.imag(E0minus_dat_2))) > self.tol:
                plt.plot(Edat, np.imag(E0minus_dat_2), "--", color="black")
        if not np.all(np.isnan(E0plus_dat_2 )):
            if np.nanmax(np.abs(np.imag(E0plus_dat_2 ))) > self.tol:
                plt.plot(Edat, np.imag(E0plus_dat_2 ), "--", color=cm(0))
        if not np.all(np.isnan(E1minus_dat_2)):
            if np.nanmax(np.abs(np.imag(E1minus_dat_2))) > self.tol:
                plt.plot(Edat, np.imag(E1minus_dat_2), "--", color=cm(1))
        if not np.all(np.isnan(E1plus_dat_2 )):
            if np.nanmax(np.abs(np.imag(E1plus_dat_2 ))) > self.tol:
                plt.plot(Edat, np.imag(E1plus_dat_2 ), "--", color=cm(2))

        plt.plot(Edat, np.real(E0minus_dat_1), color="black", label = "$E^{0-}$")
        plt.plot(Edat, np.real(E0plus_dat_1 ), color=cm(0), label = "$E^{0+}$")
        plt.plot(Edat, np.real(E1minus_dat_1), color=cm(1), label = "$E^{1-}$")
        plt.plot(Edat, np.real(E1plus_dat_1 ), color=cm(2), label = "$E^{1+}$")
        plt.plot(Edat, np.real(E0minus_dat_2), color="black")
        plt.plot(Edat, np.real(E0plus_dat_2 ), color=cm(0))
        plt.plot(Edat, np.real(E1minus_dat_2), color=cm(1))
        plt.plot(Edat, np.real(E1plus_dat_2 ), color=cm(2))



        plt.ylabel("$K^2$ (nm$^{-2}$)", size=14)
        plt.xlabel("$E$ (meV)", size=14)
        plt.legend(fontsize=14)
        plt.xlim(xlim)
        plt.ylim(ylim)
        
        if fname is not None:
            plt.savefig(fname, bbox_inches="tight")

        if return_plt:
            return plt
        else:
            plt.close()
            return


def rootsearch(f,a,b,dx):
    # Utility function used in find_levels
    # Chapter 4.3, "Numerical Methods in Engineering with Python" by Jaan Kiusalaas
    # https://stackoverflow.com/a/13055843

    x1 = a; f1 = f(a)
    x2 = a + dx; f2 = f(x2)
    while f1*f2 > 0.0:
        if x1 >= b:
            return None,None
        x1 = x2; f1 = f2
        x2 = x1 + dx; f2 = f(x2)
    return x1,x2

    
# Usage as a script 
if __name__ == "__main__":

    if len(sys.argv) > 1:
        input_fname = sys.argv[1]
    else:
        input_fname = "Rashba_exciton_options.yaml"

    RE = Rashba_Exciton(input_fname)

    # Load script options
    with open(input_fname, "r") as fin:
        input_dict = yaml.load(fin, Loader=yaml.FullLoader)

    plot_E_vs_k   = input_dict["plot_E_vs_k"]   
    plot_E_vs_ksq = input_dict["plot_E_vs_ksq"]   
    plot_ksq_vs_E = input_dict["plot_ksq_vs_E"]   
    mode          = input_dict["mode"].lower()

    if mode == "state":
        R = input_dict["R"]
        E = input_dict["E"]
        plot_wfs    = input_dict["plot_wfs"]
        include_LR     = input_dict["include_LR"]
    elif mode == "radius":
        R           = input_dict["R"]
        Emin        = input_dict["Emin"]
        Emax        = input_dict["Emax"]
        plot_wfs    = input_dict["plot_wfs"]
        save_levels = input_dict["save_levels"]
        save_osc    = input_dict["save_osc"]
        include_LR     = input_dict["include_LR"]
    elif mode == "sweep":
        Rmin        = input_dict["Rmin"]
        Rmax        = input_dict["Rmax"]
        Rstep       = input_dict["Rstep"]
        Emin        = input_dict["Emin"]
        Emax        = input_dict["Emax"]
        plot_wfs    = input_dict["plot_wfs"]
        save_levels = input_dict["save_levels"]
        save_osc    = input_dict["save_osc"]
        include_LR     = input_dict["include_LR"]

    
    if mode == "radius" or mode == "sweep":
        # Automatic Emin
        if Emin == "auto":
            Emin = RE.E_disp_min

        # Define function to find and save levels
        # Will be called once for "radius", or repeated
        # multiple times for "sweep"
        def find_and_save_levels(R):
            print(f"Calculating levels for R = {R} nm...")
            levels, wf_list = RE.find_levels(R, Emin, Emax)
            osc_list = []
            LR_list=[]
            levels_with_LR = []
            print(levels)

            if plot_wfs:
                print("Plotting wavefunctions")
                for i, wf in enumerate(wf_list):
                    RE.plot_wf(wf, R, fname=f"wf_R{R}_{i}.pdf")

            if save_levels:
                np.savetxt(f"levels_R{R}.txt", levels)

            if save_osc:
                for i, E in enumerate(levels):
                    wf = wf_list[i]
                    osc_list.append(RE.oscillator_strength(wf, R, E))
                np.savetxt(f"osc_list_R{R}.txt", osc_list)

            if include_LR:
                print("Calculating LR exhange")

                for E in levels:
                    print(f"LR exchange for E={E}")
                    wf, my_int = RE.build_wf(E, R, return_integrandLR=True)
                    try:
                        HLR = RE.LR_exch(my_int, R, E)
                    except:
                        print(f"LR integration failed for E = {E}")
                        print("Using HLR = 0")
                        HLR = 0
                    LR_list.append(HLR)
                    levels_with_LR.append(E+HLR)

                    # Save every iteration to make sure progress isn't lost if a job crashes.

                    np.savetxt(f"LR_R{R}.txt", LR_list)
                    np.savetxt(f"levels_with_LR_R{R}.txt", levels_with_LR)

            return levels, osc_list, LR_list

    if plot_E_vs_k:
        RE.plot_E_vs_k_func([-1.5,1.5], [-50,80], xlim_aug_list=[[-0.4,0.4]])

    if plot_E_vs_ksq:
        RE.plot_E_vs_ksq_func([-0.2,2.0], [-50,80], xlim_aug_list = [[-0.01, 0],[-0.05, -0.02]])

    if plot_ksq_vs_E:
        RE.plot_ksq_vs_E_func([-50, 80], [-0.5, 1.0], xlim_aug_list=[[-20,20], [RE.E_disp_min, -20]])

    if mode=="state":
        wf, my_int = RE.build_wf(E, R, return_integrandLR=True)
        if plot_wfs:
            RE.plot_wf(wf, R)

        if include_LR:
            HLR = RE.LR_exch(my_int, R, E)
            print(f"HLR = {HLR}")
            print(f"Energy with HLR = {E+HLR}")

    if mode=="radius":
        find_and_save_levels(R)


    if mode=="sweep":
        Rlist = np.arange(Rmin, Rmax+Rstep, Rstep)
        R_dat = []
        level_dat = []
        osc_dat = []
        print(f"Radius sweep from {Rmin} to {Rmax} in steps of {Rstep}")
        for R in Rlist:
            print()
            levels, osc_list, LR_list = find_and_save_levels(R)

            if include_LR:
                levels = [level[i] + LR_list[i] for i in range(len(levels))]

            R_dat += [R for level in levels]
            level_dat += levels
            osc_dat += osc_list

        osc_dat_total = np.array(osc_dat) * np.pi * np.array(R_dat)**2

        # Produce plots like Figure 3
        plt.figure(figsize=[5,5])
        if np.abs(RE.Fz) >  1:
            plt.scatter(R_dat, level_dat, c="black")
        else:
            plt.scatter(R_dat, level_dat, c=osc_dat_total, cmap=get_cmap("inferno"))
        plt.xlim([Rmin,Rmax])
        plt.title(f"$F_z = {RE.Fz}$")
        plt.ylim([Emin,Emax])
        plt.xlabel("$R$ (nm)", size=14)
        plt.ylabel("$E$ (meV)", size=14)
        plt.colorbar(label="$f$")
        plt.savefig(f"Levels_Fz{RE.Fz}.pdf", bbox_inches="tight")
        plt.close()

